from .namedLock import NamedLock as RMLock
from .namedLock import NamedLock as Lock
from .namedLock import NamedLock as RememoryLock

from .sharedBlock import RememoryBlock as RMBlock
from .sharedBlock import RememoryBlock as Block
from .sharedBlock import RememoryBlock
from .sharedBlock import BlockSize

from .sharedBool import RememoryBool as RMBool
from .sharedBool import RememoryBool as bool
from .sharedBool import RememoryBool

from .sharedDict import RememoryDict as RMDict
from .sharedDict import RememoryDict as Dict
from .sharedDict import RememoryDict

from .sharedFloat import RememoryFloat as RMFloat
from .sharedFloat import RememoryFloat as float
from .sharedFloat import RememoryFloat
from .sharedFloat import FloatTypes

from .sharedInt import RememoryInt as RMInt
from .sharedInt import RememoryInt as int
from .sharedInt import RememoryInt
from .sharedInt import IntTypes

from .sharedList import RememoryList as RMList
from .sharedList import RememoryList as List
from .sharedList import RememoryList

from .sharedString import RememoryString as RMString
from .sharedString import RememoryString as str
from .sharedString import RememoryString




from multiprocessing import Process
import base64
import pickle
from rememory import (
    RememoryDict, RememoryList, RememoryInt,
    IntTypes, RememoryFloat, RememoryString, BlockSize
)
from .worker_file import (
    worker,
    list_worker,
    int_worker,
    float_worker,
    string_worker,
    exception_worker,
    SHARED_EXCEPTION_NAME,
    SHARED_NAME,
    SHARED_LIST_NAME,
)
from .exception import NoBaseException
NUM_PROCESSES = 4

def main():
    passed = []
    failed = []

    if testSharedDict():
        passed.append("Shared Dict")
    else:
        failed.append("Shared Dict")

    if testSharedList():
        passed.append("Shared List")
    else:
        failed.append("Shared List")

    if testSharedInt():
        passed.append("Shared Int")
    else:
        failed.append("Shared Int")

    if testSharedFloat():
        passed.append("Shared Float")
    else:
        failed.append("Shared Float")

    if testSharedString():
        passed.append("Shared String")
    else:
        failed.append("Shared String")

    if testClassBlock():
        passed.append("Class Block")
    else:
        failed.append("Class Block")

    if len(failed) > 0 and len(passed) > 0:
        print("Tests Passed:")
        for test in passed:
            print(f"- {test}")
        print("Tests Failed:")
        for test in failed:
            print(f"- {test}")
    elif len(failed) > 0:
        print("All tests failed!")
    else:
        print("All tests passed!")


def testSharedDict() -> bool:
    shared: RememoryDict[str, dict[str, int]] = RememoryDict(SHARED_NAME)
    shared.clear()

    procs = [Process(target=worker) for _ in range(NUM_PROCESSES)]
    for p in procs:
        p.start()
    for p in procs:
        p.join()

    return len(shared.keys()) == NUM_PROCESSES


def testSharedList() -> bool:
    shared_list: RememoryList[str] = RememoryList(SHARED_LIST_NAME)
    with shared_list._lock:
        shared_list._write_data([])

    list_procs = [Process(target=list_worker) for _ in range(NUM_PROCESSES)]
    for p in list_procs:
        p.start()
    for p in list_procs:
        p.join()

    expected = [f"start" for _ in range(NUM_PROCESSES)] + [f"end" for _ in range(NUM_PROCESSES)]
    return list(shared_list) == expected


def testSharedInt() -> bool:
    shared_int = RememoryInt("shared_int", IntTypes.INT32)
    shared_int.value = 0

    int_procs = [Process(target=int_worker) for _ in range(NUM_PROCESSES)]
    for p in int_procs:
        p.start()
    for p in int_procs:
        p.join()

    return shared_int.value == NUM_PROCESSES


def testSharedFloat() -> bool:
    shared_float = RememoryFloat("shared_float")
    shared_float.value = 0.0

    float_procs = [Process(target=float_worker) for _ in range(NUM_PROCESSES)]
    for p in float_procs:
        p.start()
    for p in float_procs:
        p.join()

    return shared_float.value == 1.5 * NUM_PROCESSES


def testSharedString() -> bool:
    shared_string = RememoryString("shared_string", BlockSize.s128)
    shared_string.value = ""

    string_procs = [Process(target=string_worker) for _ in range(NUM_PROCESSES)]
    for p in string_procs:
        p.start()
    for p in string_procs:
        p.join()

    return shared_string.value == f"|rememory" * NUM_PROCESSES

def testClassBlock() -> bool:
    e = NoBaseException(404, "Not Found")

    raw_bytes = pickle.dumps(e)
    data_str = base64.b64encode(raw_bytes).decode("ascii")

    shared_str = RememoryString(SHARED_EXCEPTION_NAME, BlockSize.s4096)
    with shared_str._lock:
        shared_str.value = data_str

    procs = [Process(target=exception_worker) for _ in range(4)]
    for p in procs:
        p.start()
    for p in procs:
        p.join()

    reloaded = pickle.loads(base64.b64decode(shared_str.value.encode("ascii")))
    return isinstance(reloaded, NoBaseException)


import warnings
from .sharedBlock import RememoryBlock, BlockSize

class RememoryString(RememoryBlock[str]):
    """Fixed-size shared UTF-8 string built on RememoryBlock."""

    def __init__(self, name: str, size: BlockSize = BlockSize.s256):
        super().__init__(name, size.value)

    def _read_value(self) -> str:
        raw = self._read_bytes()
        return raw.rstrip(b"\x00").decode("utf-8")

    def _write_value(self, value: str):
        data = value.encode("utf-8")
        if len(data) > self._size:
            warnings.warn(
                f"RememoryString[{self._name}] value too long "
                f"({len(data)} > {self._size}), truncating."
            )
            data = data[: self._size]
        self._write_bytes(data)

    @property
    def value(self) -> str:
        with self._lock:
            return self._read_value()

    @value.setter
    def value(self, new_value: str):
        with self._lock:
            self._write_value(new_value)

    def set(self, new_value: str):
        self.value = new_value

    def get(self) -> str:
        return self.value

    def __str__(self) -> str:
        return self.value

    def __repr__(self):
        return f"<RememoryString {self._name}: '{self._read_value()}'>"



import pickle
from collections import UserList
from multiprocessing import shared_memory
from typing import Any, Generic, TypeVar, List as PyList
from .namedLock import NamedLock

T = TypeVar("T")

class RememoryList(UserList, Generic[T]):
    """
    A true shared-memory list using multiprocessing.shared_memory.
    Any Python process (spawned or launched independently) that constructs
    RememoryList("name") will attach to the same shared memory block.

    Synchronization is handled via a NamedLock (cross-process).
    """

    _value_type: Any = Any

    def __class_getitem__(cls, item):
        val_t = item

        class _TypedRememoryList(RememoryList):  # type: ignore
            _value_type = val_t

        _TypedRememoryList.__name__ = (
            f"RememoryList[{getattr(val_t, '__name__', str(val_t))}]"
        )
        return _TypedRememoryList

    def __init__(self, name: str, size: int = 65536):
        self._name = name
        self._size = size
        self._shm = None

        # Try to attach; if it doesn't exist, create it
        try:
            self._shm = shared_memory.SharedMemory(name=name)
        except FileNotFoundError:
            self._shm = shared_memory.SharedMemory(name=name, create=True, size=size)
            self._write_data([])

        self._lock = NamedLock(name)

        # Initialize UserList with a snapshot
        super().__init__(self._read_data())

    # ---------------- Internal helpers ----------------

    def _read_data(self) -> PyList[Any]:
        if self._shm is None:
            return []
        raw = bytes(self._shm.buf).rstrip(b"\x00")
        if not raw:
            return []
        try:
            return pickle.loads(raw)
        except Exception:
            return []

    def _write_data(self, data: PyList[Any]):
        if self._shm is None:
            return
        encoded = pickle.dumps(data)
        if len(encoded) > self._shm.size:
            new_size = max(len(encoded), self._shm.size * 2)
            self._shm.close()
            self._shm.unlink()
            self._shm = shared_memory.SharedMemory(
                name=self._name, create=True, size=new_size
            )
        self._shm.buf[: len(encoded)] = encoded
        self._shm.buf[len(encoded) :] = b"\x00" * (self._shm.size - len(encoded))

    # ---------------- Mutating methods ----------------

    def append(self, item: T) -> None:
        with self._lock:
            data = self._read_data()
            data.append(item)
            self._write_data(data)

    def insert(self, i: int, item: T) -> None:
        with self._lock:
            data = self._read_data()
            data.insert(i, item)
            self._write_data(data)

    def __setitem__(self, index, value):  # index can be int or slice
        with self._lock:
            data = self._read_data()
            data[index] = value
            self._write_data(data)

    def __delitem__(self, index):
        with self._lock:
            data = self._read_data()
            del data[index]
            self._write_data(data)

    def __getitem__(self, index):
        with self._lock:
            data = self._read_data()
            return data[index]

    def __len__(self) -> int:
        with self._lock:
            return len(self._read_data())

    def __iter__(self):
        with self._lock:
            return iter(self._read_data())

    def __repr__(self):
        return f"<RememoryList {self._name}: {self._read_data()}>"

    def close(self):
        if self._shm is None:
            return
        self._shm.close()

    def unlink(self):
        if self._shm is None:
            return
        self._shm.unlink()



from setuptools import setup, find_packages

setup(
    name="rememory",
    version="0.3.2",
    description="A set of lightweight shared memory variable types for safe multiprocessing access",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Nichola Walch",
    author_email="littler.compression@gmail.com",
    license="MIT",
    python_requires=">=3.8",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={
        "no": ["py.typed"],
    },
    install_requires=[
        "posix-ipc>=1.2.0; sys_platform != 'win32'",
        "pywin32>=305; sys_platform == 'win32'",
    ],
    entry_points={
        "console_scripts": [
            "rememory = testScript.__init__:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



import os
import time, pickle, base64
from rememory import RememoryDict, RememoryList, RememoryInt, IntTypes, RememoryFloat, RememoryString, BlockSize


SHARED_NAME = "game"
SHARED_LIST_NAME = "gamelist"

def worker():
    shared: RememoryDict[str, dict[str, int]] = RememoryDict(SHARED_NAME)
    pid = os.getpid()
    shared[str(pid)] = {"score": 1, "level": 1}
    time.sleep(0.5)
    temp = shared[str(pid)]
    temp["score"] += 1
    temp["level"] += 1
    shared[str(pid)] = temp


def list_worker():
    shared_list: RememoryList[str] = RememoryList(SHARED_LIST_NAME)
    pid = os.getpid()
    shared_list.append(f"start")
    time.sleep(0.5)
    shared_list.append(f"end")


def int_worker():
    shared_int = RememoryInt("shared_int", IntTypes.INT32)
    with shared_int._lock:
        shared_int.value = shared_int.value + 1


def float_worker():
    shared_float = RememoryFloat("shared_float")
    with shared_float._lock:
        shared_float.value = shared_float.value + 1.5


def string_worker():
    shared_string = RememoryString("shared_string")
    pid = os.getpid()
    with shared_string._lock:
        shared_string.value = shared_string.value + f"|rememory"

SHARED_EXCEPTION_NAME = "shared_exception"

def exception_worker():
    shared_str = RememoryString(SHARED_EXCEPTION_NAME, BlockSize.s4096)
    data = shared_str.value
    if not data:
        print(f"[{__name__}] No exception data found")
        return

    try:
        raw_bytes = base64.b64decode(data.encode("ascii"))
        exc = pickle.loads(raw_bytes)
    except Exception as e:
        print(f"[{__name__}] Failed to unpickle: {e}")